"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EggInstrumentation = exports.defaultConfig = void 0;
const api = require("@opentelemetry/api");
const instrumentation_1 = require("@opentelemetry/instrumentation");
const types_1 = require("./types");
const AttributeNames_1 = require("./enums/AttributeNames");
const version_1 = require("./version");
const utils_1 = require("./utils");
exports.defaultConfig = {
    ignoreLayers: [
        (name) => {
            return [
                'middleware - meta',
                'middleware - siteFile',
                'middleware - notfound',
                'middleware - ',
                'middleware - overrideMethod',
                'middleware - session',
                'middleware - i18n',
                'middleware - bodyParser',
            ].includes(name);
        },
    ],
};
class EggInstrumentation extends instrumentation_1.InstrumentationBase {
    constructor(config) {
        super('@opentelemetry/instrumentation-egg', version_1.VERSION, Object.assign(Object.assign({}, exports.defaultConfig), config));
    }
    init() {
        return new instrumentation_1.InstrumentationNodeModuleDefinition('egg-core', ['^4.20.0'], (moduleExports) => {
            if (moduleExports == null) {
                return moduleExports;
            }
            api.diag.debug('Patching Egg');
            const routerProto = moduleExports.EggCore.prototype;
            if (instrumentation_1.isWrapped(routerProto.use)) {
                this._unwrap(routerProto, 'use');
            }
            this._wrap(routerProto, 'use', this._getKoaUsePatch.bind(this));
            return moduleExports;
        }, (moduleExports) => {
            api.diag.debug('Unpatching Egg');
            if (instrumentation_1.isWrapped(moduleExports.EggCore.prototype.use)) {
                this._unwrap(moduleExports.EggCore.prototype, 'use');
            }
        });
    }
    _getKoaUsePatch(original) {
        const plugin = this;
        return function use(middlewareFunction) {
            let patchedFunction;
            if (middlewareFunction.router) {
                patchedFunction = plugin._patchRouterDispatch(middlewareFunction);
            }
            else {
                patchedFunction = plugin._patchLayer(middlewareFunction, false);
            }
            return original.apply(this, [patchedFunction]);
        };
    }
    _patchRouterDispatch(dispatchLayer) {
        var _a;
        api.diag.debug('Patching @koa/router dispatch');
        const router = dispatchLayer.router;
        const routesStack = (_a = router === null || router === void 0 ? void 0 : router.stack) !== null && _a !== void 0 ? _a : [];
        for (const pathLayer of routesStack) {
            const path = pathLayer.path;
            const pathStack = pathLayer.stack;
            for (let j = 0; j < pathStack.length; j++) {
                const routedMiddleware = pathStack[j];
                pathStack[j] = this._patchLayer(routedMiddleware, true, path);
            }
        }
        return dispatchLayer;
    }
    _patchLayer(middlewareLayer, isRouter, layerPath) {
        if (middlewareLayer[types_1.kLayerPatched] === true)
            return middlewareLayer;
        middlewareLayer[types_1.kLayerPatched] = true;
        api.diag.debug('patching Koa middleware layer');
        const metadata = utils_1.getMiddlewareMetadata(null, middlewareLayer, isRouter, layerPath);
        if (utils_1.isLayerIgnored(metadata.name, metadata.attributes[AttributeNames_1.AttributeNames.KOA_TYPE], this._config)) {
            api.diag.debug(`ignore Koa middleware layer ${metadata.name}`);
            return middlewareLayer;
        }
        return async (context, next) => {
            const parent = api.getSpan(api.context.active());
            if (parent === undefined) {
                return middlewareLayer(context, next);
            }
            const metadata = utils_1.getMiddlewareMetadata(context, middlewareLayer, isRouter, layerPath);
            const span = this.tracer.startSpan(metadata.name, {
                attributes: metadata.attributes,
            });
            if (!context.request.ctx.parentSpan) {
                context.request.ctx.parentSpan = parent;
            }
            if (metadata.attributes[AttributeNames_1.AttributeNames.KOA_TYPE] === types_1.KoaLayerType.ROUTER) {
                if (context.request.ctx.parentSpan.name) {
                    const parentRoute = context.request.ctx.parentSpan.name.split(' ')[1];
                    if (context._matchedRoute &&
                        !context._matchedRoute.toString().includes(parentRoute)) {
                        context.request.ctx.parentSpan.updateName(`${context.method} ${context._matchedRoute}`);
                        delete context.request.ctx.parentSpan;
                    }
                }
            }
            return api.context.with(api.setSpan(api.context.active(), span), async () => {
                try {
                    return await middlewareLayer(context, next);
                }
                catch (err) {
                    span.recordException(err);
                    throw err;
                }
                finally {
                    span.end();
                }
            });
        };
    }
}
exports.EggInstrumentation = EggInstrumentation;
EggInstrumentation.component = types_1.KoaComponentName;
//# sourceMappingURL=instrumentation.js.map